package org.netkernel.ext.system.accessor;

import java.lang.management.GarbageCollectorMXBean;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.MemoryUsage;
import java.lang.management.OperatingSystemMXBean;
import java.lang.management.ThreadMXBean;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.netkernel.container.config.IConfiguration;
import org.netkernel.container.config.IConfigurationListener;
import org.netkernel.container.impl.Kernel;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.Utils;
import org.netkernel.scheduler.RootState;

import com.ten60.netkernel.cache.se.representation2.ConcurrentCache;

public class SystemStatisticsMonitor extends Thread implements IConfigurationListener
{	
	/** housework interval */
	private long mPeriod;
	/** number of memory stats kept */
	private int mWindowSize;
	/** number of housework periods per memory stat period */
	private int mPeriodMultiplier;
	/** true if we have been told to stop */
	private boolean mStop;
	/** maximum memory **/
	private long mMax;
	
	/** the history of memory statistics */
	private long[] mStats;
	/** a smoothed baseline figure to take account of our sampling period error */
	private long mBaseLine;
	/** the last real GC baseline we captured */
	private long mRealBaseLine;
	/** a smoothed peak memory usage to take account of our sampling period error */
	private long mPeak;
	/** the last real GC peak memory we captured */
	private long mRealPeak;
	/** the last memory usage sample we took */
	private long mLast;
	/** total number of requests issued last time checked */
	private long mLastRequests;
	/** counter for implementing frequency divider */
	private int mFreqDivider;
	/** index into rolling statistics buffer **/
	private int mBufferIndex;
	
	private int mCachePeak;
	private int mCacheTrough;
	private float mReqs=10.0f;
	private int mCacheSize;
	private float mHitPercent;
	
	private static final int STAT_COUNT=11;
	
	//private Kernel mKernel;
	
	private ConcurrentCache mCache=null;
	
	public SystemStatisticsMonitor()
	{	super("System Statistics Monitor");
	}
	
	public synchronized void configurationChanged(IConfiguration aConfiguration)
	{
		long period = aConfiguration.getLong("netkernel.statistics.period", 5000L);
		int window = aConfiguration.getInt("netkernel.statistics.window", 60);
		int multiplier = aConfiguration.getInt("netkernel.statistics.multiplier", 10);
		
		if (period!=mPeriod || multiplier!=mPeriodMultiplier)
		{	//restart capture if period has changed
			mStats=null;
		}
		mPeriod=period;
		mPeriodMultiplier=multiplier;
		
		if (mWindowSize!=window || mStats==null)
		{
			if (mStats==null)
			{	mStats =  new long[window*STAT_COUNT];
				mBufferIndex=0;
			}
			else
			{	long[] stats = new long[window*STAT_COUNT];
				int j=mBufferIndex;
				int k=(window-1)*STAT_COUNT;
				for (int i=0; i<mWindowSize && i<window; i++)
				{	System.arraycopy(mStats, j, stats, k, STAT_COUNT);
					k-=STAT_COUNT;
					j-=STAT_COUNT;
					if (j<0) j=(mWindowSize-1)*STAT_COUNT;
				}
				mStats=stats;
				mBufferIndex=(window-1)*STAT_COUNT;
				
			}
			mWindowSize=window;
		}
	}	
	
	public void init(INKFRequestContext context) throws Exception
	{	
		mCache=ConcurrentCache.getInstance();
		
		MemoryPoolMXBean oldGen=Utils.getOldGen();
		MemoryUsage mu=oldGen.getUsage();
		mBaseLine=mu.getUsed();
		mLast=mBaseLine;
		mPeak=mBaseLine;
		mRealPeak=mBaseLine;
		mMax=mu.getMax()>>10;
		this.start();
	}
	
	/** stops its thread */
	public void shutdown()
	{	
		mStop=true;
		this.interrupt();
		mCache=null;
		//mKernel=null;
	}
	
	/** loops until stopped performing work and sleeping between */
	public void run()
	{	while (!mStop)
		{	try
			{	Thread.sleep(mPeriod/mPeriodMultiplier);
				this.doPeriodicHouseKeeping();
			}
			catch (Throwable e) {}
		}
	}
	
	/** calculate memory statistics by watching for GC events
	 **/
	public synchronized void doPeriodicHouseKeeping()
	{	MemoryPoolMXBean oldGen=Utils.getOldGen();
		MemoryUsage mu=oldGen.getUsage();
		long used = mu.getUsed();
		if (used<mLast)
		{	
			if (used<mRealBaseLine)
			{	mRealBaseLine = used;
			}
			else
			{	mRealBaseLine=used/2 + (mRealBaseLine/2);
			}
			if (mLast>mRealPeak)
			{	mRealPeak = mLast;
			}
			else
			{	mRealPeak=mLast/2 + (mRealPeak/2);
			}
		}
		else
		{	mRealPeak=mRealPeak*31/32+used/32;
		}
		mBaseLine = (mBaseLine*7/8) + mRealBaseLine/8;
		mPeak = (mPeak*7/8) + mRealPeak/8;
		mLast = used;
		
		
		
		mFreqDivider = (mFreqDivider+1)%mPeriodMultiplier;
		if (mFreqDivider==0)
		{	mBufferIndex = (mBufferIndex+STAT_COUNT)%(mWindowSize*STAT_COUNT);
			IHDSNode cacheStatistics=mCache.getStatistics();
			long totalReq=getRequestCount();
			float reqs;
			if (mLastRequests==0 || totalReq<mLastRequests)
			{	reqs=0.0f;
			}
			else
			{	reqs=(totalReq-mLastRequests);
			}
			mLastRequests=totalReq;
			mReqs=mReqs*0.9f + reqs*0.1f;
			
			int cachePeakSize=(Integer)cacheStatistics.getFirstValue("/cache/peakSize");
			int cacheTroughSize= (Integer)cacheStatistics.getFirstValue("/cache/troughSize");
			
			if (cachePeakSize>mCachePeak)
			{	mCachePeak=mCachePeak/2 + cachePeakSize/2;
			}
			else
			{	mCachePeak=mCachePeak*31/32+cachePeakSize/32;
			}
			if (cacheTroughSize<mCacheTrough)
			{	mCacheTrough=mCacheTrough/2+cacheTroughSize/2;
			}
			else
			{	mCacheTrough=mCacheTrough*31/32+(cacheTroughSize)/32;
			}
			
			int cacheSize=(Integer)cacheStatistics.getFirstValue("/cache/cacheSize");
			mCacheSize=mCacheSize*15/16+cacheSize/16;
			
			float hitPercent=(Float)cacheStatistics.getFirstValue("/cache/hitPercent");
			mHitPercent=mHitPercent*15/16+hitPercent/16;
			
			//cpu usage
			long now=System.currentTimeMillis();
			long period=now-mLastNow;
			long cpus=Runtime.getRuntime().availableProcessors();
			long cpuTime=0;
			
			ThreadMXBean tmxb = ManagementFactory.getThreadMXBean( );
			if (tmxb.isThreadCpuTimeSupported() && tmxb.isThreadCpuTimeEnabled())
			{
				long[] threadIds=tmxb.getAllThreadIds();
				for (int i=0; i<threadIds.length; i++)
				{	long id=threadIds[i];
					long t2 = tmxb.getThreadCpuTime( id );
					Long lid=Long.valueOf(id);
					ThreadTime tt = mThreadTimes.get(lid);
					if (t2<0)
					{	if (tt!=null)
						{	mThreadTimes.remove(lid);
						}
					}
					else if (tt==null)
					{	tt= new ThreadTime();
						tt.mCPUTime=t2;
						mThreadTimes.put(lid, tt);
					}
					else
					{	cpuTime+=t2-tt.mCPUTime;
						tt.mCPUTime=t2;
					}
				}
				
				if (mLastNow>0)
				{						
					cpuTime/=period*cpus;
					float cput=(float)cpuTime;
					mCPUTime=mCPUTime*0.9f+cput*0.1f;
				}
				
			}
			mLastNow=now;
			
			long gctime=0;
			List<GarbageCollectorMXBean> gcbeans=ManagementFactory.getGarbageCollectorMXBeans();
			for (GarbageCollectorMXBean gcbean: gcbeans)
			{	gctime+=gcbean.getCollectionTime();
			}
			long gctimeInPeriod=gctime-mLastGCTime;
			mLastGCTime=gctime;
			
			float gt=((float)gctimeInPeriod)/((float)(period*cpus));
			mGCTime=mGCTime*0.9f+gt*0.1f;
			//System.out.println("gctime="+gctimeInPeriod*1000);
			
			
			mStats[ mBufferIndex+0 ] = now;
			mStats[ mBufferIndex+1 ] = mBaseLine;
			mStats[ mBufferIndex+2 ] = mPeak;
			mStats[ mBufferIndex+3 ] = mu.getCommitted();
			mStats[ mBufferIndex+4 ] = mCachePeak;
			mStats[ mBufferIndex+5 ] = mCacheTrough;
			mStats[ mBufferIndex+6 ] = mCacheSize;
			mStats[ mBufferIndex+7 ] =(long)(mHitPercent*100.0f);
			mStats[ mBufferIndex+8 ] =(long)(mReqs*1000.0f);
			mStats[ mBufferIndex+9 ] = (long)(mCPUTime/10000);
			mStats[ mBufferIndex+10 ] = (long)(mGCTime*100);
		}
	}
	
	private long mLastNow;
	private long mLastGCTime;
	private float mGCTime;
	private float mCPUTime;
	private Map<Long,ThreadTime> mThreadTimes=new HashMap<Long,ThreadTime>();
	
	private static class ThreadTime
	{	public long mCPUTime;
	}

	
	/** write memory statistics out as IHDSNode */
	public IHDSNode getBuffer()
	{	HDSBuilder b=new HDSBuilder();
		b.pushNode("memory");
		b.addNode("max",mMax);
		//b.addNode("totalClassesLoaded",mClassLoaderBean.getTotalLoadedClassCount());
		b.addNode("totalDurationMinutes", mPeriod*mWindowSize/60000);
		b.addNode("totalDurationSeconds", mPeriod*mWindowSize/1000);
		int index = mBufferIndex;
		long maxCache=0;
		for (int i=0; i<mWindowSize; i++)
		{	index+=STAT_COUNT;
			if (index>=mWindowSize*STAT_COUNT) index=0;
			b.pushNode("stat");
				b.addNode("baseline",mStats[index+1]>>10);
				b.addNode("peak",mStats[index+2]>>10);
				b.addNode("alloc",mStats[index+3]>>10);
				Long cacheSize=(Long)mStats[index+4];
				if(cacheSize>maxCache)
				{	maxCache=cacheSize;			
				}
				b.addNode("cacheSize", mStats[index+6]);
				b.addNode("cacheSizePeak", mStats[index+4]);
				b.addNode("cacheSizeTrough", mStats[index+5]);
				b.addNode("cacheHitRate", ((float)mStats[index+7])/100.0f);
				long requests=mStats[index+8];
				float reqsec=((float)requests)/(float)mPeriod;
				float logreqsec=0;
				if(reqsec>1)
				{	logreqsec=(float)(1E6*Math.log10(reqsec));
				}
				b.addNode("logRequests", logreqsec);
				b.addNode("requests", requests/1000);
				b.addNode("time",mStats[index+0]);
				b.addNode("cpu",mStats[index+9]+mStats[index+10]);
				b.addNode("gctime",mStats[index+10]);
			b.popNode();
		}
		b.addNode("maxCache", maxCache);
		b.popNode();
		return b.getRoot();
	}
	
	/** return a filtered estimate of our GC baseline memory usage- i.e how much memory
	 * are we actually holding that cannot be released */
	public final long getBaselineMemory()
	{	return mRealBaseLine;
	}
	
	public final long getMaxMemory()
	{	return mMax<<10;
	}
	
	/** return a filtered estimate of our peak memory usage0-  i.e. how much memory are will
	 * using as a working set
	 */
	public final long getPeakMemory()
	{	return mRealPeak;
	}
	
	public final long getRequestCount()
	{	return RootState.getNextId();
	}
}